"""Defines LintableDict class."""
import typing

from ansiblelint.file_utils import Lintable


class LintableDict(dict[str, typing.Any]):
    """The LintableDict class."""

    def __init__(self, lintable: Lintable) -> None:
        """Initialize LintableDict."""
        self["base_kind"] = str(lintable.base_kind)
        self["dir"] = str(lintable.dir)
        self["exc"] = None if lintable.exc is None else str(lintable.exc)
        self["filename"] = str(lintable.filename)
        self["kind"] = str(lintable.kind)
        self["name"] = str(lintable.name)
        self["parent"] = None if lintable.parent is None else str(lintable.parent.name)
        self["role"] = str(lintable.role)
        self["stop_processing"] = bool(lintable.stop_processing)
        self["updated"] = bool(lintable.updated)



"""Execute pipelines for generating training data."""

import argparse
import logging
import os

from pathlib import Path

from .gen_ftdata import gen_ftdata_json
from .report import add_module_summary


_logger = logging.getLogger(__name__)


def run_pipeline(args: argparse.Namespace, repository_path: Path) -> int:
    """Execute pipelines for generating training data (ftdata.json)."""
    out_path = Path(args.output).absolute()
    metadata_path = out_path / "metadata"

    report_path = out_path / "report.txt"
    ftdata_path = out_path / "ftdata.json"
    lint_result_path = metadata_path / "lint-result.json"
    sage_objects_path = metadata_path / "sage-objects.json"

    if args.verbose:
        os.environ["SAGE_LOG_LEVEL"] = "debug"

    # Note: SagePipeline in imported here because of the way it initializes
    # logging, ansible-lint will generate duplicated log lines if it is
    # imported at the top-level.

    # pylint: disable=import-error,import-outside-toplevel
    from sage_scan.pipeline import (
        SagePipeline,
    )

    dp = SagePipeline()

    dp.run(
        target_dir=str(repository_path),
        lint_result=str(lint_result_path),
        output_dir=str(metadata_path),
        source={
            "data_source_description": args.source_description,
            "license": args.source_license,
            "repo_name": args.repo_name,
            "repo_url": args.repo_url,
        },
    )

    # Generate FT Data
    gen_ftdata_json(
        str(sage_objects_path),
        str(ftdata_path),
    )
    _logger.info("Training data set was created at %s.", str(ftdata_path))

    # Add module section to the report
    add_module_summary(str(sage_objects_path), args)
    _logger.info("Execution report was generated at %s.", str(report_path))

    return 0



"""Invoke ansible-lint."""
from __future__ import annotations

import sys

from pathlib import Path
from typing import TYPE_CHECKING

from ansiblelint.__main__ import (
    _do_transform,
    _logger,
    _perform_mockings_cleanup,
    get_app,
    initialize_logger,
    initialize_options,
    load_ignore_txt,
    log_entries,
    options,
    path_inject,
)
from ansiblelint.color import (
    console_options,
    reconfigure,
)


if TYPE_CHECKING:
    from ansiblelint.runner import LintResult


# pylint: disable=too-many-statements,too-many-locals
def ansiblelint_main(argv: list[str] | None = None) -> LintResult:
    """Linter CLI entry point (based on ansiblelint/__main__.py)."""
    # alter PATH if needed (venv support)
    path_inject()

    if argv is None:  # pragma: no cover
        argv = sys.argv
    initialize_options(argv[1:])

    console_options["force_terminal"] = options.colored
    reconfigure(console_options)

    initialize_logger(options.verbosity)
    for level, message in log_entries:
        _logger.log(level, message)
    _logger.debug("Options: %s", options)
    _logger.debug("CWD: %s", Path.cwd())

    # pylint: disable=import-outside-toplevel
    from ansiblelint.rules import RulesCollection
    from ansiblelint.runner import _get_matches

    app = get_app(offline=None)  # to be sure we use the offline value from settings
    rules = RulesCollection(
        options.rulesdirs,
        profile_name=options.profile,
        app=app,
        options=options,
    )

    if isinstance(options.tags, str):
        options.tags = options.tags.split(",")  # pragma: no cover
    result = _get_matches(rules, options)

    if options.write_list:
        ruamel_safe_version = "0.17.26"
        from packaging.version import Version
        from ruamel.yaml import __version__ as ruamel_yaml_version_str

        if Version(ruamel_safe_version) > Version(ruamel_yaml_version_str):
            _logger.warning(
                "We detected use of `--write` feature with a buggy ruamel-yaml %s library instead of >=%s, upgrade it before reporting any bugs like dropped comments.",
                ruamel_yaml_version_str,
                ruamel_safe_version,
            )
        _do_transform(result, options)

    mark_as_success = True

    if options.strict and result.matches:
        mark_as_success = False

    # Remove skip_list items from the result
    result.matches = [m for m in result.matches if m.tag not in app.options.skip_list]
    # Mark matches as ignored inside ignore file
    ignore_map = load_ignore_txt(options.ignore_file)
    for match in result.matches:
        if match.tag in ignore_map[match.filename]:
            match.ignored = True

    app.render_matches(result.matches)

    _perform_mockings_cleanup(app.options)
    if options.mock_filters:
        _logger.warning(
            "The following filters were mocked during the run: %s",
            ",".join(options.mock_filters),
        )

    app.report_outcome(result, mark_as_success=mark_as_success)

    return result, mark_as_success



"""Test safe checks for tar/zip files."""
import os
import random
import string
import zipfile

from pathlib import Path
from unittest import TestCase

from ansible_content_parser.safe_checks import check_zip_file_is_safe

from .test_main import sample_playbook, temp_dir


class TestSafeChecks(TestCase):
    """The TestSafeChecks class."""

    def test_with_highly_compressed_zip(self) -> None:
        """Test with a highly compressed zip file."""
        with temp_dir() as source:
            os.chdir(source.name)
            with Path("a.txt").open("w", encoding="utf-8") as f:
                f.write("A" * 4096)
            with zipfile.ZipFile(
                "test.zip",
                "w",
                compression=zipfile.ZIP_DEFLATED,
            ) as zip_file:
                zip_file.write("a.txt")
            with self.assertRaises(RuntimeError) as context:
                check_zip_file_is_safe("test.zip")

            assert (
                context.exception.args[0]
                == "ratio between compressed and uncompressed data is highly suspicious, looks like a Zip Bomb Attack"
            )

    def test_with_zip_with_too_many_files(self) -> None:
        """Test with a zip file that contains too many files."""
        with temp_dir() as source:
            os.chdir(source.name)
            with Path("0.yml").open("w", encoding="utf-8") as f:
                f.write(sample_playbook)
            with zipfile.ZipFile(
                "test.zip",
                "w",
                compression=zipfile.ZIP_DEFLATED,
            ) as zip_file:
                for i in range(10001):
                    zip_file.write(f"{i}.yml")
                    Path(f"{i}.yml").rename(f"{i+1}.yml")
            with self.assertRaises(RuntimeError) as context:
                check_zip_file_is_safe("test.zip")

            assert (
                context.exception.args[0]
                == "too many entries in this archive, can lead to inodes exhaustion of the system"
            )

    def test_with_zip_too_big(self) -> None:
        """Test with a zip file that is too big."""
        with temp_dir() as source:
            os.chdir(source.name)
            with Path("0.txt").open("w", encoding="utf-8") as f:
                f.write(sample_playbook)
                random_string = "".join(
                    random.choices(string.ascii_letters, k=1024),
                )
                f.write(random_string * 100)

            with zipfile.ZipFile(
                "test.zip",
                "w",
                compression=zipfile.ZIP_DEFLATED,
            ) as zip_file:
                for i in range(10000):
                    zip_file.write(f"{i}.txt")
                    Path(f"{i}.txt").rename(f"{i+1}.txt")
            with self.assertRaises(RuntimeError) as context:
                check_zip_file_is_safe("test.zip")

            assert (
                context.exception.args[0]
                == "the uncompressed data size is too much for the application resource capacity"
            )



"""Test __main__.py."""
import argparse
import contextlib
import json
import os
import sys
import tarfile
import zipfile

from collections.abc import Generator
from pathlib import Path
from tempfile import TemporaryDirectory
from unittest import TestCase
from unittest.mock import MagicMock, patch

import git


# pylint: disable=import-error
from ansible_content_parser.__main__ import (
    get_project_root,
    get_version,
    main,
    update_argv,
)


sample_playbook = """---
- name: Apache server installed
  hosts: web

  post_tasks:
  # Edge case: a (post) task without name
  - meta: clear_facts

  tasks:
  - name: latest Apache version installed
    yum:
      name: httpd
      state: latest

  - name: latest firewalld version installed
    yum:
      name: firewalld
      state: latest

  - name: firewalld enabled and running
    service:
      name: firewalld
      enabled: true
      state: started

  - name: firewalld permits http service
    firewalld:
      service: http
      permanent: true
      state: enabled
      immediate: yes

  # Edge case: a task with a two-line name
  - name: "Apache enabled and \
running"
    service:
      name: httpd
      enabled: true
      state: started
"""

galaxy_yml = """namespace: "namespace_name"
name: "collection_name"
version: "1.0.12"
readme: "README.md"
authors:
    - "Author1"
    - "Author2 (https://author2.example.com)"
    - "Author3 <author3@example.com>"
license:
    - "MIT"
tags:
    - demo
    - collection
repository: "https://www.github.com/my_org/my_collection"
description: Sample description
"""

# From a sample file in ansible-lint repo
sample_playbook2 = """---
- name: One
  hosts: all
  tasks:
    - name: Test when with jinja2 # noqa: jinja[spacing]
      ansible.builtin.debug:
        msg: text
      when: "{{ false }}"

- name: Two
  hosts: all
  roles:
    - role: hello
      when: "{{ '1' = '1' }}"

- name: Three
  hosts: all
  roles:
    - role: hello
      when:
        - "{{ '1' = '1' }}"
"""

# An example with a deprecated module (ec2)
sample_playbook3 = """---
- name: EC2 Sample
  hosts: all
  tasks:
    - ec2:
        key_name: mykey
        instance_type: t2.micro
        image: ami-123456
        wait: yes
        group: server
        count: 3
        vpc_subnet_id: subnet-29e63245
        assign_public_ip: yes
"""

dot_ansible_lint = """---
# .ansible-lint
profile: basic
verbosity: 1
"""

sample_playbook_name = "playbook.yml"
sample_playbook2_name = "transform-no-jinja-when.yml"
sample_playbook3_name = "ec2-sample.yml"
galaxy_yml_name = "galaxy.yml"
dot_ansible_lint_name = ".ansible-lint"
repo_name = "repo_name"


@contextlib.contextmanager
def temp_dir() -> Generator[TemporaryDirectory[str], None, None]:
    """Provide context with a temporary directory."""
    temp_directory = TemporaryDirectory()
    try:
        os.chdir(temp_directory.name)
        yield temp_directory
    finally:
        temp_directory.cleanup()


class TestMain(TestCase):
    """The TestMain class."""

    def _create_repo(self, source: TemporaryDirectory[str]) -> None:
        """Create a repository."""
        with (Path(source.name) / sample_playbook_name).open("w") as f:
            f.write(sample_playbook)
        with (Path(source.name) / galaxy_yml_name).open("w") as f:
            f.write(galaxy_yml)
        with (Path(source.name) / dot_ansible_lint_name).open("w") as f:
            f.write(dot_ansible_lint)

    def _add_second_playbook(self, source: TemporaryDirectory[str]) -> None:
        """Add the second playbook YAML file."""
        with (Path(source.name) / sample_playbook2_name).open("w") as f:
            f.write(sample_playbook2)

    def _add_third_playbook(self, source: TemporaryDirectory[str]) -> None:
        """Add the second playbook YAML file."""
        with (Path(source.name) / sample_playbook3_name).open("w") as f:
            f.write(sample_playbook3)

    def _create_tarball(
        self,
        source: TemporaryDirectory[str],
        compression: str = "",
    ) -> None:
        """Create a tarball."""
        self._create_repo(source)
        os.chdir(source.name)
        filename = (
            f"{repo_name}.tar.{compression}" if compression else f"{repo_name}.tar"
        )
        mode = f"w:{compression}" if compression else "w"
        with tarfile.open(filename, mode) as tar:
            tar.add(sample_playbook_name)
            tar.add(galaxy_yml_name)
            tar.add(dot_ansible_lint_name)

    def _create_zip_file(self, source: TemporaryDirectory[str]) -> None:
        """Create a ZIP file."""
        self._create_repo(source)
        os.chdir(source.name)
        with zipfile.ZipFile(
            f"{repo_name}.zip",
            "w",
            compression=zipfile.ZIP_DEFLATED,
        ) as zip_file:
            zip_file.write(sample_playbook_name)
            zip_file.write(galaxy_yml_name)
            zip_file.write(dot_ansible_lint_name)

    def test_cli_with_local_directory(self) -> None:
        """Run the CLI with a local directory."""
        with temp_dir() as source:
            self._create_repo(source)
            self._add_second_playbook(source)
            self._add_third_playbook(source)
            with temp_dir() as output:
                testargs = [
                    "ansible-content-parser",
                    "-v",
                    source.name + "/",  # intentionally add "/" to the end
                    output.name,
                ]
                with patch.object(sys, "argv", testargs), self.assertRaises(
                    SystemExit,
                ) as context:
                    main()

                assert context.exception.code == 0, "The exit code should be 0"

                with (Path(output.name) / "report.txt").open("r") as f:
                    for line in f:
                        if line == "Module Name     Count\n":
                            line = f.readline()
                            assert line == "---------------------\n"
                            line = f.readline()
                            assert line == "service             2\n"
                            line = f.readline()
                            assert line == "yum                 2\n"
                            line = f.readline()
                            assert line == "firewalld           1\n"
                            line = f.readline()
                            assert line == "meta                1\n"
                            line = f.readline()
                            assert line == "---------------------\n"
                            line = f.readline()
                            assert line == "TOTAL               6\n"
                            line = f.readline()
                            assert line == "---------------------\n"

    def test_cli_with_local_directory_with_no_exclude(self) -> None:
        """Run the CLI with a local directory."""
        with temp_dir() as source:
            self._create_repo(source)
            self._add_second_playbook(source)
            self._add_third_playbook(source)
            with temp_dir() as output:
                testargs = [
                    "ansible-content-parser",
                    "-v",
                    "--no-exclude",
                    source.name + "/",  # intentionally add "/" to the end
                    output.name,
                ]
                with patch.object(sys, "argv", testargs), self.assertRaises(
                    SystemExit,
                ) as context:
                    main()

                    assert context.exception.code == 1, "The exit code should be 1"

    def test_cli_with_non_archive_file(self) -> None:
        """Run the CLI with specifying a non archive file as input."""
        with temp_dir() as source:
            self._create_repo(source)
            with temp_dir() as output:
                testargs = [
                    "ansible-content-parser",
                    (Path(source.name) / sample_playbook_name).as_posix(),
                    output.name,
                ]
                with patch.object(sys, "argv", testargs):
                    with self.assertRaises(SystemExit) as context:
                        main()

                    assert context.exception.code == 1, "The exit code should be 1"

    def test_cli_with_non_existent_directory(self) -> None:
        """Run the CLI with specifying a non-existent directory as input."""
        with temp_dir() as source, temp_dir() as output:
            testargs = [
                "ansible-content-parser",
                (Path(source.name) / "non_existent_dir").as_posix(),
                output.name,
            ]
            with patch.object(sys, "argv", testargs):
                with self.assertRaises(SystemExit) as context:
                    main()

                assert context.exception.code == 1, "The exit code should be 1"

    def test_cli_with_non_empty_output_dir(self) -> None:
        """Run the CLI with a non-empty output directory."""
        with temp_dir() as source, temp_dir() as output:
            # Create a file in the output directory
            with (Path(output.name) / "b.yml").open("w") as f:
                f.write("---\nname: test\nhosts: all\n")

            testargs = ["ansible-content-parser", source.name, output.name]
            with patch.object(sys, "argv", testargs):
                with self.assertRaises(SystemExit) as context:
                    main()

                assert context.exception.code == 1, "The exit code should be 1"

    def test_cli_with_no_output_specified(self) -> None:
        """Run the CLI without specifying output."""
        with temp_dir() as source:
            testargs = ["ansible-content-parser", source.name]
            with patch.object(sys, "argv", testargs):
                with self.assertRaises(SystemExit) as context:
                    main()

                assert context.exception.code == 2, "The exit code should be 2"

    def test_cli_with_tarball(self) -> None:
        """Run the CLI with a tarball."""
        with temp_dir() as source:
            self._create_tarball(source)
            with temp_dir() as output:
                testargs = [
                    "ansible-content-parser",
                    "--profile",
                    "min",
                    (Path(source.name) / f"{repo_name}.tar").as_posix(),
                    output.name,
                ]
                with patch.object(sys, "argv", testargs):
                    with self.assertRaises(SystemExit) as context:
                        main()

                    assert context.exception.code == 0, "The exit code should be 0"

    def test_cli_with_compressed_tarball(self) -> None:
        """Run the CLI with a tarball (.tar.gz)."""
        with temp_dir() as source:
            self._create_tarball(source, "gz")
            with temp_dir() as output:
                testargs = [
                    "ansible-content-parser",
                    "--config-file",
                    source.name + "/" + dot_ansible_lint_name,
                    "--source-license",
                    "Apache",
                    "--source-description",
                    "This is a repo for test",
                    "--repo-name",
                    "test_repo",
                    "--repo-url",
                    "https://repo.example.com/test_repo",
                    str(Path(source.name) / f"{repo_name}.tar.gz"),
                    output.name,
                ]
                with patch.object(sys, "argv", testargs):
                    with self.assertRaises(SystemExit) as context:
                        main()

                    assert context.exception.code == 0, "The exit code should be 0"

                with (Path(output.name) / "ftdata.json").open("r") as f:
                    for line in f:
                        o = json.loads(line)
                        assert o["data_source_description"] == "This is a repo for test"
                        assert o["repo_name"] == "test_repo"
                        assert o["repo_url"] == "https://repo.example.com/test_repo"
                        assert o["license"] == "Apache"

    def test_cli_with_invalid_tarball(self) -> None:
        """Run the CLI with an invalid tarball."""
        with temp_dir() as source:
            self._create_zip_file(source)
            (Path(source.name) / f"{repo_name}.zip").rename(
                str(Path(source.name) / f"{repo_name}.tar"),
            )
            with temp_dir() as output:
                testargs = [
                    "ansible-content-parser",
                    str(Path(source.name) / f"{repo_name}.tar"),
                    output.name,
                ]
                with patch.object(sys, "argv", testargs):
                    with self.assertRaises(SystemExit) as context:
                        main()

                    assert context.exception.code == 1, "The exit code should be 1"

    def test_cli_with_non_existent_tarball(self) -> None:
        """Run the CLI with a non-existent tarball."""
        with temp_dir() as source, temp_dir() as output:
            testargs = [
                "ansible-content-parser",
                (Path(source.name) / f"{repo_name}.tar").as_posix(),
                output.name,
            ]
            with patch.object(sys, "argv", testargs):
                with self.assertRaises(SystemExit) as context:
                    main()

                assert context.exception.code == 1, "The exit code should be 1"

    def test_cli_with_zip_file(self) -> None:
        """Run the CLI with a zip file."""
        with temp_dir() as source:
            self._create_zip_file(source)
            with temp_dir() as output:
                testargs = [
                    "ansible-content-parser",
                    (Path(source.name) / f"{repo_name}.zip").as_posix(),
                    output.name,
                ]
                with patch.object(sys, "argv", testargs):
                    with self.assertRaises(SystemExit) as context:
                        main()

                    assert context.exception.code == 0, "The exit code should be 0"

    def test_cli_with_invalid_zip_file(self) -> None:
        """Run the CLI with an invalid zip file."""
        with temp_dir() as source:
            self._create_tarball(source, "gz")
            (Path(source.name) / f"{repo_name}.tar.gz").rename(
                str(Path(source.name) / f"{repo_name}.zip"),
            )
            with temp_dir() as output:
                testargs = [
                    "ansible-content-parser",
                    str(Path(source.name) / f"{repo_name}.zip"),
                    output.name,
                ]
                with patch.object(sys, "argv", testargs):
                    with self.assertRaises(SystemExit) as context:
                        main()

                    assert context.exception.code == 1, "The exit code should be 1"

    def test_cli_with_non_existent_zip_file(self) -> None:
        """Run the CLI with a non-existent zip file."""
        with temp_dir() as source, temp_dir() as output:
            testargs = [
                "ansible-content-parser",
                (Path(source.name) / f"{repo_name}.zip").as_posix(),
                output.name,
            ]
            with patch.object(sys, "argv", testargs):
                with self.assertRaises(SystemExit) as context:
                    main()

                assert context.exception.code == 1, "The exit code should be 1"

    @patch("git.Repo.clone_from")
    def test_cli_with_git_https_url(self, mock_clone_from: MagicMock) -> None:
        """Run the CLI with a git HTTPS URL."""
        mock_clone_from.return_value = None
        with temp_dir() as output:
            testargs = [
                "ansible-content-parser",
                "https://github.com/ansible/ansible-tower-samples.git",
                output.name,
            ]
            with patch.object(sys, "argv", testargs):
                with self.assertRaises(SystemExit) as context:
                    main()

                assert context.exception.code == 0, "The exit code should be 0"

    @patch("git.Repo.clone_from")
    def test_cli_with_git_ssh_url(self, mock_clone_from: MagicMock) -> None:
        """Run the CLI with a git SSH URL."""
        mock_clone_from.return_value = None
        with temp_dir() as output:
            testargs = [
                "ansible-content-parser",
                "git@github.com:ansible/ansible-tower-samples.git",
                output.name,
            ]
            with patch.object(sys, "argv", testargs):
                with self.assertRaises(SystemExit) as context:
                    main()

                assert context.exception.code == 0, "The exit code should be 0"

    @patch("git.Repo.clone_from")
    def test_cli_with_non_existent_git_https_url(
        self,
        mock_clone_from: MagicMock,
    ) -> None:
        """Run the CLI with a non-existent HTTPS git URL."""
        mock_clone_from.side_effect = git.GitCommandError("clone")
        with temp_dir() as output:
            testargs = [
                "ansible-content-parser",
                "https://github.com/ansible/non-existent-samples.git",
                output.name,
            ]
            with patch.object(sys, "argv", testargs):
                with self.assertRaises(SystemExit) as context:
                    main()

                assert context.exception.code == 1, "The exit code should be 1"

    def test_update_argv(self) -> None:
        """Test __main__.update_argv()."""
        input_data = {
            "skip_transform": True,
            "verbose": True,
            "config_file": "config.file",
            "profile": "basic",
        }
        args = argparse.Namespace(**input_data)
        argv = ["__DUMMY__"]
        update_argv(argv, args)

        assert "--fix=all" not in argv
        assert "-v" in argv
        assert "--config-file" in argv
        assert "config.file" in argv
        assert "--profile" in argv
        assert "basic" in argv

        input_data = {
            "skip_transform": False,
            "verbose": False,
            "config_file": None,
            "profile": None,
        }
        args = argparse.Namespace(**input_data)
        argv = ["__DUMMY__"]
        update_argv(argv, args)

        assert "--fix=all" in argv
        assert "-v" not in argv
        assert "--config-file" not in argv
        assert "--profile" not in argv

    def test_get_project_root(self) -> None:
        with temp_dir() as output:
            output_path = Path(output.name)
            repository_path = output_path / "repository"
            repository_path.mkdir()

            # When there is no file/directory in repository_path,
            # get_project_root return repository_path
            path = get_project_root(repository_path)
            assert path.name == repository_path.name

            # When there is one single directory in repository_path,
            # get_project_root return the directory path
            project_path = repository_path / "project"
            project_path.mkdir()
            path = get_project_root(repository_path)
            assert path.name == project_path.name

            # When there are more than one directories in repository_path,
            # get_project_root return repository_path
            project_path2 = repository_path / "project2"
            project_path2.mkdir()
            path = get_project_root(repository_path)
            assert path.name == repository_path.name

    def test_get_version(self) -> None:
        version_str = get_version()

        assert "ansible-content-parser" in version_str
        assert "ansible-lint" in version_str
        assert "ansible-core" in version_str
        assert "(not found)" not in version_str



"""Check if a given archive (zip/tar) file is safe."""

# Ref: https://sonarcloud.io/organizations/ansible/rules?open=python%3AS5042&rule_key=python%3AS5042&tab=how_to_fix
import tarfile
import zipfile


threshold_entries = 10000
threshold_size = 1000000000

# Increased threshold_ratio from 10 to 100 for reduce false positives.
threshold_ratio = 100


def _check_total_size_and_entries(
    total_entry_archive: int,
    total_size_archive: int,
) -> None:
    if total_entry_archive > threshold_entries:
        msg = "too many entries in this archive, can lead to inodes exhaustion of the system"
        raise RuntimeError(
            msg,
        )
    if total_size_archive > threshold_size:
        msg = "the uncompressed data size is too much for the application resource capacity"
        raise RuntimeError(
            msg,
        )


def check_tar_file_is_safe(source: str) -> None:
    """Make sure that expanding the tar file is safe."""
    if not tarfile.is_tarfile(source):
        msg = f"{source} is not a valid tar archive file."
        raise RuntimeError(
            msg,
        )

    total_size_archive = 0
    total_entry_archive = 0

    with tarfile.open(source) as f:  # NOSONAR
        for entry in f:
            tarinfo = f.extractfile(entry)
            if tarinfo:
                total_entry_archive += 1
                size_entry = 0
                while True:
                    size_entry += 1024
                    total_size_archive += 1024

                    # Commented out following lines because they do not seem to detect highly compressed tar (e.g .tar.gz
                    # files) like the corresponding logic for zip files.  We should be OK because we will check total
                    # size and entries later.

                    # ratio = size_entry / entry.size
                    # if ratio > threshold_ratio:
                    #     msg = "ratio between compressed and uncompressed data is highly suspicious, looks like a Zip Bomb Attack"
                    #     raise RuntimeError(
                    #         msg,
                    #     )

                    chunk = tarinfo.read(1024)
                    if not chunk:
                        break

            _check_total_size_and_entries(total_entry_archive, total_size_archive)


def check_zip_file_is_safe(source: str) -> None:
    """Make sure that expanding the zip file is safe."""
    if not zipfile.is_zipfile(source):
        msg = f"{source} is not a valid zip file."
        raise RuntimeError(
            msg,
        )

    total_size_archive = 0
    total_entry_archive = 0

    with zipfile.ZipFile(source, "r") as f:
        for info in f.infolist():
            data = f.read(info)
            total_entry_archive += 1

            total_size_archive = total_size_archive + len(data)
            ratio = (len(data) / info.compress_size) if info.compress_size != 0 else 1
            if ratio > threshold_ratio:
                msg = "ratio between compressed and uncompressed data is highly suspicious, looks like a Zip Bomb Attack"
                raise RuntimeError(
                    msg,
                )

            _check_total_size_and_entries(total_entry_archive, total_size_archive)



"""Generate reports."""
import argparse
import datetime
import json
import logging

from collections import Counter
from pathlib import Path
from typing import TypedDict

from sarif import loader  # pylint: disable=import-error
from sarif.operations import summary_op  # pylint: disable=import-error

from .lintable_dict import LintableDict
from .version import __version__


_logger = logging.getLogger(__name__)

_label_count = "Count"
_label_file_type = "File Type"
_label_file_path = "File Path"
_label_file_state = "Updated"
_label_module_name = "Module Name"
_label_total = "TOTAL"

_report_txt = "report.txt"


def filetype_summary(result: dict[str, list[LintableDict]]) -> str:
    """Calculate summary stats for file types."""
    kinds = {f["filename"]: f["kind"] for f in result["files"]}
    counts = Counter(kinds.values())
    entries = []
    total = 0
    max_kind_len = len(_label_file_type)
    max_count_len = 0
    for kind in sorted(counts, key=lambda x: counts[x], reverse=True):
        kind_str = "(file type not identified)" if kind == "" else kind
        count = str(counts[kind])
        total += counts[kind]
        entries.append([kind_str, count])
        if len(kind_str) > max_kind_len:
            max_kind_len = len(kind_str)

    max_count_len = max(len(str(total)), len(_label_count))

    num_spaces = 5
    separator = "-" * (max_kind_len + num_spaces + max_count_len)
    summary = separator + "\n"
    summary += (
        _label_file_type.ljust(max_kind_len)
        + " " * num_spaces
        + _label_count.rjust(max_count_len)
        + "\n"
    )
    summary += separator + "\n"
    for kind_str, count in entries:
        summary += (
            kind_str.ljust(max_kind_len)
            + " " * num_spaces
            + count.rjust(max_count_len)
            + "\n"
        )
    summary += separator + "\n"
    summary += (
        _label_total.ljust(max_kind_len)
        + " " * num_spaces
        + str(total).rjust(max_count_len)
        + "\n"
    )
    summary += separator

    return summary


def get_file_list_summary(files: list[LintableDict]) -> str:
    """Get summary string from the lintable list."""
    entries = []
    max_filename_len = len(_label_file_path)
    max_kind_len = len(_label_file_type)
    max_state_len = len(_label_file_state)
    kinds = {f["filename"]: f["kind"] for f in files}
    updated = {f["filename"]: f["updated"] for f in files}
    for filename in sorted(kinds):
        kind = kinds[filename]
        if kind != "":  # Skip files that was not identified by ansible-lint
            state = "updated" if updated[filename] else ""
            entries.append([filename, kind, state])
            if len(filename) > max_filename_len:
                max_filename_len = len(filename)
            if len(kind) > max_kind_len:
                max_kind_len = len(kind)
            if len(state) > max_state_len:
                max_state_len = len(state)

    num_spaces = 2
    separator = "-" * (
        max_filename_len + num_spaces + max_kind_len + num_spaces + max_state_len
    )
    summary = separator + "\n"
    summary += (
        _label_file_path.ljust(max_filename_len)
        + " " * num_spaces
        + _label_file_type.ljust(max_kind_len)
        + " " * num_spaces
        + _label_file_state
        + "\n"
    )
    summary += separator + "\n"
    for filename, kind, state in entries:
        summary += (
            filename.ljust(max_filename_len)
            + " " * num_spaces
            + kind.ljust(max_kind_len)
            + " " * num_spaces
            + state
            + "\n"
        )
    summary += separator

    return summary


def get_sarif_summary(metadata_path: Path, sarif_file: str) -> str:
    """Get summary string from the SARIF JSON file."""
    sarif_summary = metadata_path / "sarif_summary.txt"
    sarif_file_set = loader.load_sarif_files(sarif_file)
    summary_op.generate_summary(sarif_file_set, str(sarif_summary), False)
    with sarif_summary.open(encoding="utf-8") as f:
        summary = f.read()
    sarif_summary.unlink()
    return summary


class _SageObject(TypedDict):
    annotations: dict[str, str]
    module: str
    module_info: dict[str, str]


def get_module_name(o: _SageObject) -> str:
    """Find a module name in a dictionary that represents an Ansible task."""
    module_name = o.get("module", "")
    if not module_name:
        _logger.error("No module name was found in a task: %s", json.dumps(o))
        module_name = "(not found)"
    return module_name


def get_module_summary(sage_objects: str) -> str:
    """Get summary string for Ansible modules found in running the Content Parser."""
    modules = []
    with Path(sage_objects).open(encoding="utf-8") as f:
        for line in f:
            o = json.loads(line)
            if o.get("py/object") == "sage_scan.models.Task":
                modules.append(get_module_name(o))

    counts = Counter(modules)
    entries = []
    max_module_len = len(_label_module_name)
    total = 0
    for module in sorted(counts, key=lambda x: (-counts[x], x)):
        count = str(counts[module])
        total += counts[module]
        entries.append([module, count])
        if len(module) > max_module_len:
            max_module_len = len(module)

    max_count_len = max(len(str(total)), len(_label_count))

    num_spaces = 5
    separator = "-" * (max_module_len + num_spaces + max_count_len)
    summary = separator + "\n"
    summary += (
        _label_module_name.ljust(max_module_len)
        + _label_count.rjust(max_count_len + num_spaces)
        + "\n"
    )
    summary += separator + "\n"
    for module, count in entries:
        summary += (
            module.ljust(max_module_len)
            + " " * num_spaces
            + count.rjust(max_count_len)
            + "\n"
        )
    summary += separator + "\n"
    summary += (
        _label_total.ljust(max_module_len)
        + " " * num_spaces
        + str(total).rjust(max_count_len)
        + "\n"
    )
    summary += separator
    return summary


def get_excluded_files(excluded: list[str]) -> str:
    """Get the list of excluded files in the second ansible-lint run."""
    max_filename_len = len(_label_file_path)
    for file_path in excluded:
        if len(file_path) > max_filename_len:
            max_filename_len = len(file_path)

    summary = "-" * max_filename_len + "\n"
    summary += _label_file_path + "\n"
    summary += "-" * max_filename_len + "\n"
    for file_path in excluded:
        summary += f"{file_path}\n"
    summary += "-" * max_filename_len + "\n"

    return summary


def generate_report(
    json_file: str,
    sarif_file: str,
    sarif_file2: str,
    args: argparse.Namespace,
) -> None:
    """Generate report."""
    report = f"""
********************************************************************************
****                Ansible Content Parser Execution Report                 ****
********************************************************************************

Date/Time             : {datetime.datetime.now(tz=datetime.timezone.utc).isoformat()}
Content Parser Version: {__version__}
Source Repository     : {args.source}
Output Directory      : {args.output}
"""
    out_path = Path(args.output)
    metadata_path = out_path / "metadata"

    if json_file:
        with Path(json_file).open(encoding="utf-8") as f:
            result = json.load(f)
            files = result["files"]
            excluded = result.get("excluded", [])

        report += f"""

[ File counts per type ]

{filetype_summary(result)}


[ List of Ansible files identified ]

{get_file_list_summary(files)}


[ Issues found by ansible-lint ]
"""
    if sarif_file2:
        report += f"""
(First run)
{get_sarif_summary(metadata_path, sarif_file)}

(Second run)

- Files excluded from the second run due to syntax-check errors found in the first run

{get_excluded_files(excluded)}
{get_sarif_summary(metadata_path, sarif_file2)}
"""
    else:
        report += f"""
{get_sarif_summary(metadata_path, sarif_file)}
"""
    with (out_path / _report_txt).open(mode="w") as f:
        f.write(report)

    return


def add_module_summary(sage_objects: str, args: argparse.Namespace) -> None:
    """Add Ansible module summary to the report."""
    summary = get_module_summary(sage_objects)

    out_path = Path(args.output)
    report_path = out_path / _report_txt

    with report_path.open(mode="a") as f:
        f.write(
            f"""

[ List of Ansible modules found in tasks ]

{summary}
""",
        )



"""A runpy entry point for ansible-content-parser.

This makes it possible to invoke CLI
via :command:`python -m ansible_content_parser`.
"""
import argparse
import contextlib
import errno
import json
import logging
import os
import shutil
import sys
import tarfile
import typing
import zipfile

from collections.abc import Generator
from importlib.metadata import PackageNotFoundError, version
from pathlib import Path

import giturlparse  # pylint: disable=import-error

from ansiblelint.constants import RC
from git import Repo
from packaging.version import Version

from .lint import ansiblelint_main
from .lintable_dict import LintableDict
from .pipeline import run_pipeline
from .report import generate_report
from .safe_checks import check_tar_file_is_safe, check_zip_file_is_safe
from .version import __version__


_logger = logging.getLogger(__name__)


# https://stackoverflow.com/questions/6194499/pushd-through-os-system
@contextlib.contextmanager
def pushd(new_dir: str) -> Generator[None, None, None]:
    """Use this in a with block to run some operations from a given directory."""
    previous_dir = Path.cwd()
    os.chdir(new_dir)
    try:
        yield
    finally:
        os.chdir(previous_dir)


def execute_ansiblelint(
    argv: list[str],
    work_dir: str,
) -> dict[str, typing.Any]:
    """Execute ansible-lint."""
    with pushd(work_dir):
        # Clear root logger handlers as ansible-lint adds one without checking existing ones.
        logging.getLogger().handlers.clear()

        result, mark_as_success = ansiblelint_main(argv)
        return {
            "files": [LintableDict(lintable) for lintable in result.files],
        }


def parse_args(argv: list[str]) -> argparse.Namespace:
    """Parse arguments."""
    parser = argparse.ArgumentParser(
        description="Parse Ansible files in the given repository by running ansible-lint and generate"
        " a training dataset for Ansible Lightspeed.",
    )
    parser.add_argument(
        "--config-file",
        help=" Specify the configuration file to use for ansible-lint.  "
        "By default it will look for '.ansible-lint', '.config/ansible-lint.yml', or '.config/ansible-lint.yaml' "
        "in the source repository.",
    )
    parser.add_argument(
        "--profile",
        choices=["min", "basic", "moderate", "safety", "shared", "production"],
        help="Specify which rules profile to be used for ansible-lint",
    )
    parser.add_argument(
        "--skip-transform",
        action="store_true",
        help="Skip the transform step of ansible-lint.  If this option is not specified, ansible-lint is executed "
        "with the --fix option and files are transformed according to the rules specified.",
    )
    parser.add_argument(
        "--skip-ansible-lint",
        action="store_true",
        help="Skip the execution of ansible-lint.",
    )
    parser.add_argument(
        "--no-exclude",
        action="store_true",
        help="Do not rerun ansible-lint with excluding files that caused syntax check errors. If one or more syntax "
        "check errors were found, execution fails without generating the training dataset.",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="Explain what is being done",
    )
    parser.add_argument(
        "--source-license",
        default="",
        help="Specify the license that will be included in the training dataset.",
    )
    parser.add_argument(
        "--source-description",
        default="",
        help="Specify the description of the source that will be included in the training dataset.",
    )
    parser.add_argument(
        "--repo-name",
        default="",
        help="Specify the repository name that will be included in the training dataset. "
        "If it is not specified, it is generated from the source name.",
    )
    parser.add_argument(
        "--repo-url",
        default="",
        help="Specify the repository url that will be included in the training dataset. "
        "If it is not specified, it is generated from the source name.",
    )
    parser.add_argument(
        "--version",
        action="version",
        version=get_version(),
    )
    parser.add_argument(
        "source",
        help="source, which can be an zip/tar archive, a git URL or a local directory",
    )
    parser.add_argument(
        "output",
        help="output directory",
    )
    args = parser.parse_args(argv)
    args.output = str(Path(args.output).absolute())

    if args.config_file:
        args.config_file = str(Path(args.config_file).absolute())

    return args


def set_repo_name_and_repo_url(args: argparse.Namespace, is_file: bool) -> None:
    """Set repository name and URL from source name when they are not set explicitly."""
    if not args.repo_name:
        repo_name = args.source
        if repo_name.endswith("/"):
            repo_name = repo_name[:-1]
        i = repo_name.rfind("/")
        if i != -1:
            repo_name = repo_name[i + 1 :]
        i = repo_name.find(".")
        if i != -1:
            repo_name = repo_name[:i]
        args.repo_name = repo_name

    if not args.repo_url:
        args.repo_url = (
            Path(args.source).absolute().as_uri() if is_file else args.source
        )


def prepare_source_and_output(args: argparse.Namespace) -> Path:
    """Prepare source (archive/url/directory) and output directory."""
    source, output = args.source, args.output

    supported_tar_file_extensions = [
        ".tar",
        ".tar.gz",
        ".tgz",
        ".tar.bz2",
        ".tbz2",
        ".tar.xz",
        ".txz",
    ]

    out_path = setup_output(output)
    repository_path = out_path / "repository"
    metadata_path = out_path / "metadata"

    # Check if the specified source is a supported archive.
    if source.endswith(".zip"):
        try:
            check_zip_file_is_safe(source)
            with zipfile.ZipFile(source) as zip_file:
                repository_path.mkdir()
                zip_file.extractall(repository_path)
                metadata_path.mkdir()
                set_repo_name_and_repo_url(args, True)
                return get_project_root(repository_path)
        except Exception:
            _logger.exception(
                "An exception thrown in extracting files from %s.",
                source,
            )
            sys.exit(1)
    else:
        for ext in supported_tar_file_extensions:
            if source.endswith(ext):
                try:
                    check_tar_file_is_safe(source)
                    with tarfile.open(source) as tar:  # NOSONAR
                        repository_path.mkdir()
                        tar.extractall(repository_path)
                        metadata_path.mkdir()
                        set_repo_name_and_repo_url(args, True)
                    return get_project_root(repository_path)
                except Exception:
                    _logger.exception(
                        "An exception thrown in extracting files from %s.",
                        source,
                    )
                    sys.exit(1)

    # Check if the specified source is a URL
    if giturlparse.validate(source):
        try:
            repository_path.mkdir()
            Repo.clone_from(source, repository_path)
            metadata_path.mkdir()
            set_repo_name_and_repo_url(args, False)
            return repository_path
        except Exception:
            _logger.exception(
                "An exception thrown in cloning git repository %s.",
                source,
            )
            sys.exit(1)

    # Assume the source is a local directory
    if Path(source).is_dir():
        # As shutil.copytree creates repository_path, we do not need to call repository_path,mkdir()
        shutil.copytree(source, repository_path)
        metadata_path.mkdir()
        set_repo_name_and_repo_url(args, True)
    else:
        _logger.error("%s is not a directory.", source)
        sys.exit(1)

    return repository_path


def setup_output(output: str) -> Path:
    """Set up output directory."""
    # Check if the specified output directory exists.
    out_path = Path(output)
    if not out_path.is_dir():
        out_path.mkdir(parents=True)

    # Make sure the directory is empty.
    files = os.listdir(output)
    if len(files) > 0:
        _logger.error("Output directory is not empty.")
        sys.exit(1)

    return out_path


def get_project_root(repository_path: Path) -> Path:
    """Get the project root directory from the given path."""
    files = os.listdir(repository_path)
    # If the given path contains a single directory at root assume it as the project root directory.
    if len(files) == 1:
        path = Path(repository_path / files[0])
        return path if path.is_dir() else repository_path
    return repository_path


def main() -> None:
    """Invoke the main entrypoint with current CLI args.

    This function also processes the runtime exceptions.
    """
    return_code = 0
    try:
        args = parse_args(sys.argv[1:])
        repository_path = prepare_source_and_output(args)

        out_path = Path(args.output)
        metadata_path = out_path / "metadata"

        sarif_file = str(metadata_path / "sarif.json")
        argv = ["__DUMMY__", "--sarif-file", sarif_file]
        update_argv(argv, args)

        try:
            execute_lint_step(args, argv, metadata_path, repository_path, sarif_file)
        except Exception:
            _logger.exception("An exception was thrown while running ansible-lint.")
            sys.exit(1)

        return_code = run_pipeline(args, repository_path)

    except OSError as exc:
        # NOTE: Only "broken pipe" is acceptable to ignore
        if exc.errno != errno.EPIPE:  # pragma: no cover
            raise
    except KeyboardInterrupt:  # pragma: no cover
        _logger.info("Terminated by a keyboard interrupt.")
        sys.exit(RC.EXIT_CONTROL_C)
    except RuntimeError:  # pragma: no cover
        raise

    sys.exit(return_code)


def get_version() -> str:
    """Return version string that contains versions of important dependents."""
    msg = f"ansible-content-parser {__version__} using"
    for k in ["ansible-lint", "ansible-core"]:
        try:
            v = Version(version(k))
            msg += f" {k}:{v}"
        except PackageNotFoundError:
            msg += f" {k}:(not found)"
    return msg


def execute_lint_step(
    args: argparse.Namespace,
    argv: list[str],
    metadata_path: Path,
    repository_path: Path,
    sarif_file: str,
) -> None:
    """Execute ansible-lint and create metadata files."""
    exclude_paths: list[str] = []
    if not args.skip_ansible_lint:
        serializable_result = execute_ansiblelint(
            argv,
            str(repository_path),
        )
        lint_result = str(metadata_path / "lint-result.json")
        with Path(lint_result).open(
            "w",
            encoding="utf-8",
        ) as f:
            f.write(json.dumps(serializable_result))

        parse_sarif_json(exclude_paths, sarif_file)

        # If syntax-errors occurred on some files, kick off the second run excluding those files
        if len(exclude_paths) > 0 and not args.no_exclude:
            lint_result2 = str(metadata_path / "lint-result-2.json")
            sarif_file2 = str(metadata_path / "sarif-2.json")
            argv = ["__DUMMY__", "--sarif-file", sarif_file2]
            argv.append("--exclude")
            argv.extend(exclude_paths)
            update_argv(argv, args)
            _logger.info(",".join(argv))
            serializable_result_2 = execute_ansiblelint(
                argv,
                str(repository_path),
            )
            serializable_result_2["excluded"] = exclude_paths

            with Path(lint_result2).open(mode="w", encoding="utf-8") as f:
                f.write(json.dumps(serializable_result_2))
        else:
            lint_result2 = ""
            sarif_file2 = ""

    generate_report(
        "" if args.skip_ansible_lint else lint_result2 if lint_result2 else lint_result,
        sarif_file,
        sarif_file2,
        args,
    )

    if len(exclude_paths) > 0 and args.no_exclude:
        msg = "One or more syntax-check errors were found by ansible-lint"
        raise RuntimeError(msg)


def parse_sarif_json(exclude_paths: list[str], sarif_file: str) -> None:
    """Analyze SARIF.json to see if syntax-check errors occurred or not on the first run."""
    with Path(sarif_file).open("rb") as f:
        o = json.load(f)
        for run in o["runs"]:
            for result in run["results"]:
                if result["ruleId"].startswith("syntax-check"):
                    exclude_paths.extend(
                        [
                            location["physicalLocation"]["artifactLocation"]["uri"]
                            for location in result["locations"]
                        ],
                    )


def update_argv(argv: list[str], args: argparse.Namespace) -> None:
    """Update arguments to ansible-lint based on arguments given to ansible-content-parser."""
    if not args.skip_transform:
        argv.append("--fix=all")
    if args.verbose:
        argv.append("-v")
    if args.config_file:
        argv.append("--config-file")
        argv.append(args.config_file)
    if args.profile:
        argv.append("--profile")
        argv.append(args.profile)


if __name__ == "__main__":
    main()



"""Generate FT data."""
import json
import logging

from pathlib import Path
from typing import TypedDict

# pylint: disable=import-error
from sage_scan.models import (
    SageObject,
    Task,
    load_objects,
)
from sage_scan.process.utils import get_tasks


_logger = logging.getLogger(__name__)


class _FTRecord(TypedDict):
    data_source_description: str
    input: str
    license: str
    output: str
    path: str
    repo_name: str
    repo_url: str


class _FTDataError(Exception):
    pass


def _gen_ftdata(task: Task, parent: SageObject) -> _FTRecord:
    record = _FTRecord(
        data_source_description="",
        input="",
        license="",
        output="",
        path="",
        repo_name="",
        repo_url="",
    )

    record["data_source_description"] = task.source.get("data_source_description", "")
    record["license"] = task.source.get("license", "")
    record["repo_url"] = task.source.get("repo_url", "")
    record["repo_name"] = task.source.get("repo_name", "")
    record["path"] = task.filepath

    output = ""
    task_name = task.name
    yaml_lines = task.yaml_lines

    # split on the task name to determine the output
    if task_name:
        separator = task_name
        if task_name not in yaml_lines and task_name[-1] == ".":
            separator = task_name[:-1]
        parts = yaml_lines.split(separator)
        if len(parts) >= 2:
            # Remove the trailing spaces and newline characters from name value
            # if they are present after splitting on just the task name
            output = parts[1].lstrip(" ").lstrip("\n")

    # If we don't have an output, then we don't have a valid sample
    if output == "":
        msg = "No output is found."
        raise _FTDataError(msg)

    record["output"] = output

    parent_yaml = parent.yaml_lines

    # Get everything up to and including the prompt as the input
    # if its the first task then
    _input = ""
    task_line_num = task.line_num_in_file[0]
    if task_line_num:
        _input = "\n".join(parent_yaml.splitlines()[:task_line_num])

    # If we don't have an input, then we don't have a valid sample
    if _input == "":
        msg = "No input is found."
        raise _FTDataError(msg)

    record["input"] = _input

    return record


def gen_ftdata_json(sage_objects_json: str, ftdata_json: str) -> None:
    """Generate ftdata.json file."""
    record_lines = []
    for project in load_objects(sage_objects_json).projects():
        parents = []
        if project.playbooks:
            parents.extend(project.playbooks)
        if project.taskfiles:
            parents.extend(project.taskfiles)
        for parent in parents:
            for task in get_tasks(root=parent, project=project):
                try:
                    record = _gen_ftdata(task=task, parent=parent)
                    record_lines.append(json.dumps(record) + "\n")
                except _FTDataError:
                    _logger.debug(
                        "No FT Data was generated for the task:\n"
                        "  (name): %s\n"
                        "  (yaml_lines): %s",
                        task.name,
                        task.yaml_lines,
                    )

    with Path(ftdata_json).open("w", encoding="utf-8") as file:
        file.write("".join(record_lines))
